#!/bin/bash
# 
# Version 1.0 - Mark Swift
#
# Run this script from the directory that requires to be accessed by the QLAY emulator.
# It recognises existing 'qlay.dir' files, Qemulator/QPC file headers and XTcc (data length) file entries.
# The script creates a 'qlay.dir' containing a 64-byte QDOS file header for each file.

# Return the integer filesize in bytes
if_FILESZ() # filename
{  
  if [ "${sv_OS}" == "Darwin" ]
  then
    stat -f %z -- "${1}"
  else
    stat --printf %s -- "${1}"
  fi
}

# Return the integer file modification date as epoch
if_FILEMODDATEEPOCH() # filename
{  
  if [ "${sv_OS}" == "Darwin" ]
  then
    stat -f %m -- "${1}"
  else
    stat --printf %Y -- "${1}"
  fi
}

# Return the file modification date as CCYYMMDDhhmm.SS
if_FILEMODDATESTRING() # filename
{  
  if [ "${sv_OS}" == "Darwin" ]
  then
    stat -f %Sm -t %Y%m%d%H%M.%S -- "${1}"
  else
    date -d @$(stat --printf %Y -- "${1}") +"%Y%m%d%H%M.%S"
  fi
}

# Integer function to return the fileoffset of a given hexstring in the file filename
if_FINDF_H() # filename,hexstring
{ 
  iv_Pos=$(xxd -ps -u -c 0 -- "${1}" | tr -d "\n" | grep -b -o ${2} | cut -d":" -f1 | head -n1)
  if test -n "${iv_Pos}"
  then
    printf $((${iv_Pos}/2))
  fi
}

# Integer function to return the big endian long word (4 byte) value stored at fileoffset in the file filename
if_PEEKF_L() # filename,fileoffset
{  
  echo $((16#$(dd if="${1}" ibs=1 skip=${2} count=4 status=none | xxd -p)))
}

# Integer function to return the big endian short word (2 byte) value stored at fileoffset in the file filename
if_PEEKF_W() # filename,fileoffset
{  
  echo $((16#$(dd if="${1}" ibs=1 skip=${2} count=2 status=none | xxd -p)))
}

# Integer function to return the byte value stored at fileoffset in the file filename
if_PEEKF_B() # filename,fileoffset
{  
  echo $((16#$(dd if="${1}" ibs=1 skip=${2} count=1 status=none | xxd -p)))
}

# Null function to poke the big endian long word (4 byte) value into the file filename at location fileoffset
nf_POKEF_L() # filename,fileoffset,value
{  
  printf "%08X\n" ${3} | xxd -r -p | dd of="${1}" bs=1 seek=${2} conv=notrunc status=none
}

# Null function to poke the big endian short word (2 byte) value into the file filename at location fileoffset
nf_POKEF_W() # filename,fileoffset,value
{  
  printf "%04X\n" ${3} | xxd -r -p | dd of="${1}" bs=1 seek=${2} conv=notrunc status=none
}

# Null function to poke a byte value into the file filename at location fileoffset
nf_POKEF_B() # filename,fileoffset,value
{  
  printf "%02X\n" ${3} | xxd -r -p | dd of="${1}" bs=1 seek=${2} conv=notrunc status=none
}

# Null function to poke the string value into the file filename at location fileoffset
nf_POKEF_S() # filename,fileoffset,value
{  
  printf -- "${3}" | xxd -ps -u -c 0 | xxd -r -p | dd of="${1}" bs=1 seek=${2} conv=notrunc status=none
}

# Full path of this script
sv_ThisScriptFilePath="${0}"
  
# Get dir of this script
sv_ThisScriptDirPath="$(dirname "${sv_ThisScriptFilePath}")"

# Change dir to where this script is located
cd "${sv_ThisScriptDirPath}"

# Get the os name
sv_OS="$(uname -s)"

# Get the architecture 
sv_Arch="$(uname -p)"

# Check that we are not accidentally running from the wrong directory
if ! test -e "IAGREE"
then
  echo
  echo "qlayt.sh processes all the files in a directory as follows:"
  echo
  echo "1. Replaces '.' in file names and replaces with '_'"
  echo "2. Strips Qemulator/QPC file headers from files"
  echo "3. Creates or replaces the 'qlay.dir' file"
  echo
  echo "It would be quite inconvenient if you ran it from the wrong directory, so..."
  echo "Please create a file called 'IAGREE' to allow the script to process a directory,"
  echo
  exit 0
fi

# The difference between QL zero epoch and Unix zero epoch is 3287 days
iv_LinuxEpochQLdate=$((3287*86400))

sv_QemuHdrHex=$(printf ']!QDOS File Header' | xxd -ps -u -c 0 | tr -d "\n" )
sv_XTccHex=$(printf 'XTcc' | xxd -ps -u -c 0 | tr -d "\n" )
sv_QlayDirFileName="qlay.dir"
sv_TmpQlayDirFileName="${sv_QlayDirFileName}.tmp"
iv_offs=0

rm -f ${sv_TmpQlayDirFileName}

while read sv_FileName
do
  iv_FileNameLength=${#sv_FileName}
  if [ ${iv_FileNameLength} -lt 36 ]
  then
    iv_FileFileTyp=0
    iv_FileDataLen=0

    iv_FileSize=$(($(if_FILESZ "${sv_FileName}")+64))
    iv_FileModQLdate=$(($(if_FILEMODDATEEPOCH "${sv_FileName}")+${iv_LinuxEpochQLdate}))
    iv_FileModString=$(if_FILEMODDATESTRING "${sv_FileName}")

    # Change file ._ext to _ext
    sv_NewFileName=$(echo ${sv_FileName} | sed "s|\.\([^._]*$\)|_\1|")
    if [ "${sv_NewFileName}" != "${sv_FileName}" ]
    then
      mv -- "${sv_FileName}" "${sv_NewFileName}"
      touch -t ${iv_FileModString} -- "${sv_NewFileName}"
      sv_FileName="${sv_NewFileName}"
    fi

    # Check any existing qlay.dir for the file entry
    if test -e ${sv_QlayDirFileName}
    then
      sv_FileNameHex=$(printf "%04X" ${iv_FileNameLength})$(printf -- ${sv_FileName} | xxd -ps -u -c 0 | tr -d "\n" )
      iv_Pos=$(if_FINDF_H "${sv_QlayDirFileName}" "${sv_FileNameHex}")

      if test -n "${iv_Pos}"
      then
        # Get the file filetype/datalen
        iv_Pos=$((64*$((${iv_Pos}/64))))
        iv_FileFileTyp=$(if_PEEKF_B "${sv_QlayDirFileName}" $((${iv_Pos}+5)))
        iv_FileDataLen=$(if_PEEKF_L "${sv_QlayDirFileName}" $((${iv_Pos}+6)))
 
        if [ ${iv_FileFileTyp} -ne 0 ]
        then
          echo "* Found Qlaydir:${iv_FileFileTyp},${iv_FileDataLen}"
        fi
     fi
    fi

    # Check for a Qemulator/QPC file header
    iv_Pos=$(if_FINDF_H "${sv_FileName}" "${sv_QemuHdrHex}")

    if test -n "${iv_Pos}"
    then
      if [ ${iv_Pos} -eq 0 ]
      then
        iv_QemuHdrWordLen=$(if_PEEKF_B "${sv_FileName}" 19)
        iv_QemuHdrByteLen=$((${iv_QemuHdrWordLen}+${iv_QemuHdrWordLen}))

        iv_FileFileTyp=$(if_PEEKF_B "${sv_FileName}" 21)
        iv_FileDataLen=$(if_PEEKF_L "${sv_FileName}" 22)

        # Strip off the file header
        dd if="${sv_FileName}" of="${sv_FileName}.${iv_FileModString}" ibs=1 skip=${iv_QemuHdrByteLen} status=none
        touch -t ${iv_FileModString} "${sv_FileName}.${iv_FileModString}"
        rm -f "${sv_FileName}"
        mv "${sv_FileName}.${iv_FileModString}" "${sv_FileName}"

        echo "* Found QemuHdr:${iv_QemuHdrWordLen},${iv_FileFileTyp},${iv_FileDataLen}"
      fi
    fi

    # Check for an XTcc (data length) file entry
    iv_Pos=$(if_FINDF_H "${sv_FileName}" "${sv_XTccHex}")

    if test -n "${iv_Pos}"
    then
      iv_Pos=$((${iv_Pos}+4))
      iv_FileFileTyp=1
      iv_FileDataLen=$(if_PEEKF_L "${sv_FileName}" ${iv_Pos})
      echo "* Found XTcc:${iv_Pos},${iv_FileDataLen}"
    fi

    # Write info to the 'qlay.dir' 64 byte file header entry for the file
    nf_POKEF_L "${sv_TmpQlayDirFileName}" $((${iv_offs}+0)) ${iv_FileSize}
    nf_POKEF_W "${sv_TmpQlayDirFileName}" $((${iv_offs}+14)) ${iv_FileNameLength}
    nf_POKEF_S "${sv_TmpQlayDirFileName}" $((${iv_offs}+16)) ${sv_FileName}
    nf_POKEF_L "${sv_TmpQlayDirFileName}" $((${iv_offs}+52)) ${iv_FileModQLdate}

    if [ ${iv_FileFileTyp} -ne 0 ]
    then
      nf_POKEF_B "${sv_TmpQlayDirFileName}" $((${iv_offs}+5)) ${iv_FileFileTyp}
      nf_POKEF_L "${sv_TmpQlayDirFileName}" $((${iv_offs}+6)) ${iv_FileDataLen}
    fi

    nf_POKEF_L "${sv_TmpQlayDirFileName}" $((${iv_offs}+60)) 0

    echo ${sv_FileName}
    iv_offs=$((${iv_offs}+64))
  fi
done < <(find . -type f -prune \! \(  -iname "*.info" -or -name "*.DS_Store" -or -iname "*.uaem" -or -iname "*.-UQLX-" -or -iname "*_UAEFSDB.___" -or -iname "qlay.dir" -or -iname "qlayt.sh" -or -iname "IAGREE" \) -exec echo {} \; | cut -d"/" -f2 )

# Replace any existing 'qlay.dir' file with the new one
rm -f ${sv_QlayDirFileName}
mv ${sv_TmpQlayDirFileName} ${sv_QlayDirFileName}

# KEY - just in case you wonder why the variabless have strange names
#    bv_  - string variable with the values "true" or "false"
#    iv_  - integer variable
#    sv_  - string variable
#
#    nf_  - null function    (doesn't return a value)
#    bf_  - boolean function (returns string values "true" or "false"
#    if_  - integer function (returns an integer value)
#    sf_  - string function  (returns a string value)

